clear; clc;
load fcst_draws;
t_start = t + 1;

%% load data
loaddata;
[Ycore, Yothers] = standardize_data( Ycore, Yothers );

%% MCMC for individual models setup
nsims = 20000;          % retained MCMC draws
burnin = 0.1 * nsims;	% burnin period            
simstep = 20;           % draws to skip for thinning
M = 0500;               % number of IS draws for ML computation
svsims = floor( nsims / simstep );

%% MCMC for composite posterior resampling of non-core parameters
cpnsims = nsims / 2;    % make sure cpnsims / cpsimstep = svsims (otherwise bad things happen)
cpsimstep = simstep / 2;

%% fixed specification parameters
loff = 4;   % VAR lag offset (i.e., first VAR lag is y(t-loff), etc.)
p = 4;      % order of the VAR
nother = 1; % number of variables from Yothers to include in each submodel;
            % note: the current code version must have nother = 1 and
            % will not function properly if this setting is changed

[T, ncore] = size( Ycore );         % dimensions of core variables
n = ncore + nother;                 % number of variables in a submodel
m_all = size( Yothers, 2 );         % total number of submodels
T = T - loff - p + 1;               % adjust the time dimension for initial values
dates = dates( p + loff:end );      % adjust dates according to the new time dimension
kbeta = n * n * p + n;              % no. of parameter in beta
kgam = n * ( n - 1 ) / 2;           % no. of parameter in gamma_t

%% other algorithm settings
DIC = 1;                % compute BIC and DIC if DIC = 1
ML = 1;                 % compute ML if ML = 1
ttmax = 1;              % maximum forecast horizon
startingdate = 1970;    % starting date for forecasting (default: 1970Q1)

%% construct data objects
[Y_all, X_all, W_all] = construct_data( Ycore, Yothers, T, n, p, loff, kgam, m_all );
stn = struct( 'idx', [], 'n', ncore, 'p', p );  % to enforce stationary on the VAR in core
                                                % variables only embeded in each sub-model

%% initialize the forecasting exercise
T0 = find( dates == startingdate );
simresults = struct( 'beta', cell( m_all, 1 ), ...
                     'gam', cell( m_all, 1 ), 'h', cell( m_all, 1 ), ...
                     'Siggam', cell( m_all, 1 ), 'Sigh', cell( m_all, 1 ), ...
                     'gam0', cell( m_all, 1 ), 'h0', cell( m_all, 1 ), ...
                     'mlx', cell( m_all, 1 ), 'mlw', cell( m_all, 1 ), ...
                     'dic', cell( m_all, 1 ) );

% initialize the random number generator
rng( 'shuffle' ); rngstate = rng; % a new state
% load rngstate-001; rng( rngstate ); % a previously saved state

% initialize parallel processing and the text-based progress monitor
doparallel = setup_parpool();
[txtlen, ~] = setup_progmons( false );

disp( ['Running recursive forecasting exercise for ' num2str( T - loff + 1 - T0 ) ' periods...'] );
runtime = 0; tic; % start the clock
for t = t_start:T - loff + 1 - T0
    Tt = T0 + t - 1;

    % useful quantities for computing conditional MLs and combining draws
    coreY = repmat( [true( ncore, 1 ); false( nother, 1 )], Tt, 1 );
    coreX = [repmat( [true; repmat( [true( ncore, 1 ); false( nother, 1 )], p, 1 )], ncore, 1 ); ...
            false( nother * ( 1 + n * p ), 1 )];
    otherX = [false( ncore * ( 1 + n * p ), 1 ); true( nother * ( 1 + n * p ), 1 )];
    coreW = repmat( [true( ncore, 1 ); false( kgam - ncore, 1 )], Tt, 1 );
    stn.idx = coreX; % to enforce stationarity of embedded VAR in core vars

    %% construct priors
    [beta0, Vbeta, agam, Vgam, ah, Vh, nugam0, Sgam0, nuh0, Sh0, prior_all, prior_core] = ...
        construct_priors( Y_all, X_all, coreY, otherX, coreW, n, Tt, p, kbeta, kgam, m_all );

    % set up the java-based progress monitor
    yr = floor( dates( Tt ) );
    qt = 1 + ( dates( Tt ) - yr ) / .25;
    heading = ['Running MCMC for ' num2str( yr ) 'Q' num2str( qt ) ': '];
    [~, ppm] = setup_progmons( doparallel, heading, m_all );

    %% estimate all sub-models in parallel
    parfor mi = 1:m_all
        % the following need to be predefined for parfor construction
        mlx = []; mlx_std = [];
        mlw = []; mlw_std = [];
        dic = []; pD = []; Dbar = []; maxllike = [];
        
        % estimate a sub-model
        yt = reshape( Y_all{mi}( 1:Tt, : )', [], 1 );
        Xt = SURform2( X_all{mi}( 1:Tt, : ), n );
        Wt = W_all{mi}( 1:Tt * n, 1:Tt * kgam );
        
        [s_beta, s_gam, s_h, s_Siggam, s_Sigh, s_gam0, s_h0] ...
            = VARSV_mcmc( yt, Xt, Wt, Tt, n, stn, kbeta, kgam, ...
                beta0, Vbeta, agam, Vgam, ah, Vh, nugam0, Sgam0, nuh0, Sh0, ...
                nsims, burnin, svsims, simstep );

        %% compute MLs for evaluating sub-model likelihoods
        [mlx, mlx_std] ...
            = ml_tvp_r1_sv_core( yt( coreY ), s_beta( :, ~otherX ), ...
            s_Siggam( :, coreW( 1:kgam ) ), s_Sigh( :, coreY( 1:n ) ), ...
            s_h0( :, coreY( 1:n ) ), s_gam0( :, coreW( 1:kgam ) ), ...
            prior_core, Xt( coreY, ~otherX ), Wt( coreY, coreW ), ncore, M );

        %% compute MLs for composite likelihood weights
        if ML == 1
            % marginal likelihood for only the ncore variables in a VAR-SV sub-model
            [mlw, mlw_std] ...
                = ml_tvp_r1_sv_core( yt, s_beta, s_Siggam, s_Sigh, s_h0, s_gam0, ...
                prior_all, Xt, Wt, ncore, M );
        end

        %% compute BIC and DIC for composite likelihood weights
        if DIC == 1
            % DIC for all (ncore+1) variables in a VAR-SV sub-model
            [dic, pD, Dbar, maxllike] ...
                = dic_tvp_r1_sv_core( yt, s_beta, s_Siggam, s_Sigh, s_h0, s_gam0, ...
                Xt, Wt, ncore, ceil( svsims / M ) );
        end
        
        % save sim results for processing after completing all sub-model computation
        simresults( mi ).beta = s_beta;
        simresults( mi ).gam = s_gam;
        simresults( mi ).h = s_h;
        simresults( mi ).Siggam = s_Siggam;
        simresults( mi ).Sigh = s_Sigh;
        simresults( mi ).gam0 = s_gam0;
        simresults( mi ).h0 = s_h0;
        simresults( mi ).mlx = [mlx mlx_std];
        simresults( mi ).mlw = [mlw mlw_std];
        simresults( mi ).dic = [dic pD Dbar maxllike];

        %% generate one-period ahead forecasts from individual sub-model
        fcst_all{mi, t} = VARSV_fcst( Y_all{mi}( Tt + loff:Tt + loff - 1 + ttmax, : ), X_all{mi}( Tt + loff, : ), ...
                            s_beta, s_gam, s_h, s_Siggam, s_Sigh, ncore, kgam, svsims, loff );

        % update the progress monitor
        if ~isempty( ppm )
            ppm.increment();
        end
    end
    
    % close the progress window if it exists
    if ~isempty( ppm )
        ppm.delete();
    end

    % compute CL weights
    [wCL, GA_pden] = comp_CL_weights( simresults, fcst_all( :, t ), GA_pden );
    
    % generate forecasts from the GA optimal pool
    fY = Y_all{1}( Tt + loff:Tt + loff - 1 + ttmax, 1:ncore );
    fcst_GA = 0;
    for mi = 1:m_all
        fcst_GA = fcst_GA + wCL{4}( mi ) * fcst_all{mi, t};
    end
	yerr = bsxfun( @minus, fcst_GA( :, :, 1:ncore, 1 ), reshape( fY, [1 1 3] ) );
    fcst_GA( :, :, 1:ncore, 3 ) = yerr .^ 2;
    fcst_GA( :, :, 1:ncore, 4 ) = abs( yerr ) ...
        - .5 * abs( fcst_GA( :, :, 1:ncore, 1 ) - fcst_GA( :, :, 1:ncore, 2 ) );
    fcst_all{end - 1, t} = fcst_GA;

    % pool draws from individual submodels to obtain draws from the
    % composite posterior (i.e. derived from the composite likelihood)
    % and generate forecasts based on pooled predictive likelihoods

	fX1 = X_all{1}( Tt + loff, coreX( 1:1 + n * p ) );
    fX2 = zeros( 1, nother * p, m_all );
    for mi = 1:m_all
        fX2( :, :, mi ) = X_all{mi}( Tt + loff, ~coreX( 1:1 + n * p ) );
    end
    
    s_lnqCL = [];
    fcst_CL = cell( length( wCL ), 1 );
    
    for j = 1:length( wCL )
        % resample core parameters using simple rejection sampling
        [b_core, g_core, h_core, sg_core, sh_core, s_lnqCL] ...
                        = VARCL_rs( Y_all, X_all, W_all, beta0, Vbeta, ...
                        simresults, wCL{j}, coreY, coreX, otherX, coreW, ...
                        kgam, Tt, n, ncore, svsims, doparallel, s_lnqCL );

%         %%%%% TO BE COMPLETED %%%%%
%         % resample noncore parameters using MCMC
        kgam_other = nnz( ~coreW( 1:kgam ) );
        b_other = nan( svsims, nnz( otherX ), m_all );
        g_other = nan( svsims, nnz( ~coreW ), m_all );
        h_other = nan( svsims, nother * Tt, m_all );
        sg_other = nan( svsims, kgam_other, m_all );
        sh_other = nan( svsims, nother, m_all );
%         
%         [~, ppm] = setup_progmons( doparallel, 'Running MCMC for noncore params: ', m_all );
%         wCLj = wCL{j};
%         
%         parfor mi = 1:m_all
%             if wCLj( mi ) > 1 / ( m_all * 1e6 )
%                 % estimate a submodel if it has a nontrivial weight
%                 yt = reshape( Y_all{mi}( 1:Tt, : )', [], 1 );
%                 Xt = SURform2( X_all{mi}( 1:Tt, : ), n );
%                 Wt = W_all{mi}( 1:Tt * n, 1:Tt * kgam );
% 
%                 % start with draws from sub-model MCMC
%                 startvals = struct( ...
%                     'beta', mean( simresults( mi ).beta( :, otherX ) )', ...
%                     'gam', mean( simresults( mi ).gam( :, ~coreW ) )', ...
%                     'h', reshape( mean( simresults( mi ).h( :, ~coreY ) ), nother, Tt )', ...
%                     'Siggam', mean( simresults( mi ).Siggam( :, ~coreW( 1:kgam ) ) )', ...
%                     'Sigh', mean( simresults( mi ).Sigh( :, ~coreY( 1:n ) ) )', ...
%                     'gam0', mean( simresults( mi ).gam0( :, ~coreW( 1:kgam ) ) )', ...
%                     'h0', mean( simresults( mi ).h0( :, ~coreY( 1:n ) ) )' );
% 
%                 [b_other( :, :, mi ), g_other( :, :, mi ), h_other( :, :, mi ), ...
%                 sg_other( :, :, mi ), sh_other( :, :, mi )] ...
%                     = VARSV_mcmc( yt( ~coreY ), Xt( ~coreY, otherX ), Wt( ~coreY, ~coreW ), ...
%                         Tt, nother, [], nnz( otherX ), nnz( ~coreW( 1:kgam ) ), ...
%                         beta0( otherX ), Vbeta( otherX ), ...
%                         agam( ~coreW( 1:kgam ) ), Vgam( ~coreW( 1:kgam ) ), ...
%                         ah( ~coreY( 1:n ) ), Vh( ~coreY( 1:n ) ), ...
%                         nugam0( ~coreW( 1:kgam ) ), Sgam0( ~coreW( 1:kgam ) ), ...
%                         nuh0( ~coreY( 1:n ) ), Sh0( ~coreY( 1:n ) ), ...
%                         cpnsims, 0, svsims, cpsimstep, startvals, log( wCLj( mi ) ) );
% 
%                 % update the progress monitor
%                 if ~isempty( ppm )
%                     ppm.increment();
%                 end
%             end
%         end
% 
%         % close the progress window if it exists
%         if ~isempty( ppm )
%             ppm.delete();
%         end
        %%%%%%%%%%%%%%%%%%%%%%%%%%%

        % generate forecasts from the composite predictive posteriors
        fcst_CL{j} = VARCL_fcst( fY, fX1, fX2, b_core, g_core, h_core, sg_core, sh_core, ...
                          b_other, g_other, h_other, sg_other, sh_other, ...
                          wCL{j}, beta0( ~coreX & ~otherX ), diag( Vbeta( ~coreX & ~otherX ) ), ...
                          ncore, nother, kgam - kgam_other, kgam_other, m_all, svsims, loff );
    end
    fcst_all{end, t} = fcst_CL;

    % save current draws (in case something bad happens)
    save -v7.3 fcst_draws fcst_all GA_pden t;

    % show progress
    [runtime, txtlen] = showprog( t, T - loff + 1 - T0, T - loff + 1 - T0, runtime, toc, txtlen );
end
fprintf( '\nForecasting completed after %5.3f hours.\n', toc / 3600 );